(** Common types, functions, and data structures for the CoAlgebraic reasoner.
    @author Florian Widmann
 *)


(*****************************************************************************)
(*     Types that can be modified rather easily                              *)
(*     (of course you then have to modify the implementations below too)     *)
(*****************************************************************************)

(* This type has to be extended for additional logics. *)
type functors =
  | MultiModalK
  | MultiModalKD
  | CoalitionLogic
  | GML
  | Choice
  | Fusion

(* This type may be extended for additional logics. *)
type formulaType =
  | Other
  | ConstF
  | AndF
  | OrF
  | AtF
  | NomF
  | ExF
  | AxF
  | EnforcesF
  | AllowsF
  | MoreThanF
  | MaxExceptF
  | ChcF
  | FusF

type localFormula
type bset

type atFormula
type cset
type csetSet

type lht
type fht
type nht

type state

type core

type setState

type setCore

type setCnstr


(*****************************************************************************)
(*                Types that are hard coded into the algorithm               *)
(*****************************************************************************)

and stateExpander = unit -> ruleEnumeration

and sort = CoAlgFormula.sort

and nodeStatus =
  | Expandable
  | Open
  | Sat
  | Unsat

and dependencies = bset list -> bset
and ruleEnumeration =
  | AllInOne of (dependencies * (sort * bset) list) list
  | NextRule of dependencies * (sort * bset) list
  | NoMoreRules

type nominal = sort * localFormula

type node =
  | State of state
  | Core of core

type constrnt =
  | UnsatC
  | SatC
  | OpenC of node list
  | UnexpandedC of node list

type propagateElement =
  | UState of state * int option
  | UCore of core * bool
  | UCnstr of cset

type queueElement =
  | QState of state
  | QCore of core
  | QCnstr of cset
  | QEmpty

type sortTable = (functors * int list) array

exception CoAlg_finished of bool


val sortTable : sortTable ref


(*****************************************************************************)
(*      "Module type" and a specific implementation of the main queue        *)
(*****************************************************************************)

val queueInit : unit -> unit
val queueIsEmpty : unit -> bool
val queueInsertState : state -> unit
val queueInsertCore : core -> unit
val queueInsertCnstr : cset -> unit
val queueGetElement : unit -> queueElement

val doNominalPropagation : unit -> bool
val doSatPropagation : unit -> bool


(*****************************************************************************)
(*        "Module type" and a specific implementation of the graph           *)
(*****************************************************************************)

val graphInit : unit -> unit
val graphIterStates : (state -> unit) -> unit
val graphIterCores : (core -> unit) -> unit
val graphIterCnstrs : (cset -> constrnt -> unit) -> unit
val graphClearCnstr : unit -> unit
val graphFindState : sort -> bset -> state option
val graphFindCore : sort -> bset -> core option
val graphFindCnstr : cset -> constrnt option
val graphInsertState : sort -> bset -> state -> unit
val graphInsertCore : sort -> bset -> core -> unit
val graphInsertCnstr : cset -> constrnt -> unit
val graphReplaceCnstr : cset -> constrnt -> unit
val graphSizeState : unit -> int
val graphSizeCore : unit -> int
val graphSizeCnstr : unit -> int
val graphAddRoot : core -> unit
val graphGetRoot : unit -> core


(*****************************************************************************)
(*     "Module type" and a specific implementation of sets of constraints    *)
(*****************************************************************************)

val cssEmpty : csetSet
val cssExists : (cset -> bool) -> csetSet -> bool
val cssForall : (cset -> bool) -> csetSet -> bool
val cssUnion : csetSet -> csetSet -> csetSet
val cssAdd : cset -> csetSet -> csetSet
val cssFold : (cset -> 'a -> 'a) -> csetSet -> 'a -> 'a
val cssSingleton : cset -> csetSet
val cssEqual : csetSet -> csetSet -> bool


(*****************************************************************************)
(*      "Module type" and a specific implementation of state nodes           *)
(*****************************************************************************)

val stateMake : sort -> bset -> stateExpander -> state
val stateGetSort : state -> sort
val stateGetBs : state -> bset
val stateSetBs : state -> bset -> unit
val stateGetStatus : state -> nodeStatus
val stateSetStatus : state -> nodeStatus -> unit
val stateGetParents : state -> core list
val stateAddParent : state -> core -> unit
val stateGetRule : state -> int -> dependencies * core list
val stateGetRules : state -> (dependencies * core list) list
val stateAddRule : state -> dependencies -> core list -> int
val stateGetConstraints : state -> csetSet
val stateSetConstraints : state -> csetSet -> unit
val stateNextRule : state -> ruleEnumeration


(*****************************************************************************)
(*      "Module type" and a specific implementation of core nodes            *)
(*****************************************************************************)

val coreMake : sort -> bset -> Minisat.solver -> fht -> core
val coreGetSort : core -> sort
val coreGetBs : core -> bset
val coreSetBs :core -> bset -> unit
val coreGetStatus : core -> nodeStatus
val coreSetStatus : core -> nodeStatus -> unit
val coreGetParents :core -> (state * int) list
val coreAddParent : core -> state -> int -> unit
val coreGetChildren : core -> state list
val coreAddChild : core -> state -> unit
val coreGetConstraints : core -> csetSet
val coreSetConstraints : core -> csetSet -> unit
val coreGetSolver : core -> Minisat.solver
val coreGetFht : core -> fht
val coreGetConstraintParents : core -> cset list
val coreAddConstraintParent : core -> cset -> unit


(*****************************************************************************)
(*      "Module type" and a specific implementation of sets of               *)
(*      states, cores, and constraints for the sat propagation               *)
(*****************************************************************************)

val setEmptyState : unit -> setState
val setEmptyCore : unit -> setCore
val setEmptyCnstr : unit -> setCnstr
val setAddState : setState -> state -> unit
val setAddCore : setCore -> core -> unit
val setAddCnstr : setCnstr -> cset -> unit
val setMemState : setState -> state -> bool
val setMemCore : setCore -> core -> bool
val setMemCnstr : setCnstr -> cset -> bool
val setRemoveState : setState -> state -> unit
val setRemoveCore : setCore -> core -> unit
val setRemoveCnstr : setCnstr -> cset -> unit
val setIterState : (state -> unit) -> setState -> unit
val setIterCore : (core -> unit) -> setCore -> unit
val setIterCnstr : (cset -> unit) -> setCnstr -> unit


(*****************************************************************************)
(*        "Module type" and a specific implementation of the mapping         *)
(*        between (local) formulae and literals of the sat solver,           *)
(*        and of a mapping of nominals to sets of formulae                   *)
(*****************************************************************************)

val lhtInit : unit -> lht
val lhtAdd : lht -> Minisat.lit -> localFormula -> unit
val lhtFind : lht -> Minisat.lit -> localFormula option

val fhtInit : unit -> fht
val fhtAdd : fht -> localFormula -> Minisat.lit -> unit
val fhtFind : fht -> localFormula -> Minisat.lit option

val nhtInit : unit -> nht
val nhtAdd : nht -> nominal -> bset -> unit
val nhtFind : nht -> nominal -> bset option
val nhtFold : (nominal -> bset -> 'a -> 'a) -> nht -> 'a -> 'a


(*****************************************************************************)
(*         "Module type" and a specific implementation of sets of            *)
(*         local formulae and @-formulae                                     *)
(*****************************************************************************)

val bsetMake : unit -> bset (* a new bset which only contains True *)
val bsetMakeRealEmpty : unit -> bset (* a new bset without containing True *)
val bsetAdd : bset -> localFormula -> unit
val bsetMem : bset -> localFormula -> bool
val bsetFold : (localFormula -> 'a -> 'a) -> bset -> 'a -> 'a
val bsetIter : (localFormula -> unit) -> bset -> unit
val bsetFilter : bset -> (localFormula -> bool) -> bset
val bsetAddTBox : sort -> bset -> bset
val bsetCopy : bset -> bset

val csetMake : unit -> cset
val csetAdd : cset -> atFormula -> unit
val csetCopy : cset -> cset
val csetUnion : cset -> cset -> cset
val csetHasDot : cset -> bool
val csetAddDot : cset -> unit
val csetRemDot : cset -> unit
val csetAddTBox : sort -> cset -> bset
val csetIter : cset -> (atFormula -> unit) -> unit


(*****************************************************************************)
(*            "Module type" and a specific implementation of                 *)
(*            local formulae and @-formulae                                  *)
(*****************************************************************************)

val lfGetType : sort -> localFormula -> formulaType
val lfGetDest1 : sort -> localFormula -> localFormula
val lfGetDest2 : sort -> localFormula -> localFormula
val lfGetDest3 : sort -> localFormula -> int
val lfGetDestNum : sort -> localFormula -> int
val lfGetDestAg : sort -> localFormula -> int array
val lfGetNeg : sort -> localFormula -> localFormula option
val lfGetAt : nominal -> localFormula -> atFormula
val lfToInt : localFormula -> int
val lfFromInt : int -> localFormula

val atFormulaGetSubformula : atFormula -> localFormula
val atFormulaGetNominal : atFormula -> nominal

val lfToAt : sort -> localFormula -> atFormula

val ppFormulae : (string -> sort option) -> CoAlgFormula.sortedFormula list -> CoAlgFormula.sortedFormula ->
    CoAlgFormula.sortedFormula list * CoAlgFormula.sortedFormula * bset
