
open CoolUtils
module L = List
module CoAlg = CoAlgFormula
module R = CoAlgReasoner
module CM = CoAlgMisc

exception ConversionError of string

(* Just a copy'n'paste from the *.mli file *)
type uri = string
type role =
  | ROLE of uri
  | INVERSE of role

type class_exp =
  | THING
  | NOTHING
  | NAMED   of uri
  | NOT     of class_exp
  | AND     of class_exp list
  | OR      of class_exp list
  | SOME    of role * class_exp
  | ALL     of role * class_exp
  | MIN     of int * role * class_exp
  | MAX     of int * role * class_exp
  (* TODO: | OneOf of  * formula *)

type declaration =
  | CLASS of uri
  | DATATYPE of uri
  | OBJECTPROPERTY of uri
  | DATAPROPERTY of uri
  | ANNOTATIONPROPERTY of uri
  | NAMEDINDIVIDUAL of uri


type axiom =
  | ANNOTATION of string
  (* class expression axioms *)
  | DECLARATION of declaration
  | SUBCLASS of class_exp * class_exp
  | EQUIVALENTCLASSES of class_exp list
  | DISJOINTCLASSES of class_exp list
  | DISJOINTUNION   of class_exp * (class_exp list)

type prefix = string * string (* prefix name to long substitution *)
type prefixes = prefix list
type ontology = string * (axiom list) (* ontology uri and axioms *)


(* actual content *)
let dummy = "remove this as soon there are any useful declarations"


let precedence ce = match ce with
    | THING | NOTHING -> 0
    | NAMED _ -> 0
    | NOT _ -> 1
    | SOME (_,_) | ALL  (_,_)  -> 2
    | MIN (_,_,_)     | MAX  (_,_,_) -> 2
    | AND _       | OR  _    -> 3

let precedence_role r = match r with
    | ROLE _ -> 0
    | INVERSE _ -> 1

let string_of_role r =
    let rec sor oldpred r =
        let p = precedence_role r in
        (if p >= oldpred then "(" else "")^(
        match r with
        | ROLE str  -> str
        | INVERSE r -> sor p r ^ "¯"
        )^(if p >= oldpred then ")" else "")
    in
    sor 10 r

let string_of_class_exp ce =
    let rec soc oldpred ce =
        let p = precedence ce in
        (if p >= oldpred then "(" else "")^(
        let s = soc p in (* something like "show" *)
        match ce with
            | THING     -> "⊤"
            | NOTHING   -> "⊥"
            | NAMED u   -> u
            | NOT c -> "¬"^(s c)
            | AND la -> String.concat " ⊓ " (L.map s la)
            | OR  la -> String.concat " ⊔ " (L.map s la)
            | SOME (r,c) -> "∃"^(string_of_role r)^"."^(s c)
            | ALL  (r,c) -> "∀"^(string_of_role r)^"."^(s c)
            | MIN  (n,r,c) -> "≥"^(string_of_role r)^"."^(s c)
            | MAX  (n,r,c) -> "≤"^(string_of_role r)^"."^(s c)
        )^(if p >= oldpred then ")" else "")
    in
    soc 10 ce

let string_of_axiom ax =
    let soc = string_of_class_exp in
    match ax with
    | ANNOTATION str -> "»"^str^"«"
    | DECLARATION a -> "Declaration (TODO)"
    | SUBCLASS (a,b) -> (soc a) ^" ⊑ "^(soc b)
    | EQUIVALENTCLASSES la -> String.concat " ≡ " (L.map soc la)
    | DISJOINTCLASSES la -> String.concat " ⊍ " (L.map soc la)
    | DISJOINTUNION (union,parts) -> (soc union)^" ≡ "^(String.concat " ⊍ " (L.map soc parts))

let string_of_prefix (short,u) = "Prefix "^short^" ↦ "^u

let string_of_ontology (name,axioms) =
    "Ontology " ^ name ^ " {\n"^(
    String.concat "\n" (L.map (((^) "  ") << string_of_axiom) axioms)
    )^"\n}\n"

let rec coalg_formula_of_owl (ce:class_exp): CoAlgFormula.formula =
    let o2c = coalg_formula_of_owl in
    let o2r r = match r with
                | ROLE str -> str
                | INVERSE _ -> raise (ConversionError "Can not express inverses with coalgebraic formulas")
    in
    match ce with
    | THING -> CoAlg.TRUE
    | NOTHING -> CoAlg.FALSE
    | NAMED str -> CoAlg.AP str
    | NOT ce -> CoAlg.NOT (o2c ce)
    | AND l -> L.fold_left (fun x y -> CoAlg.AND (x, o2c y)) CoAlg.TRUE l
    | OR l -> L.fold_left (fun x y -> CoAlg.OR (x, o2c y)) CoAlg.FALSE l
    | SOME (r,c) -> CoAlg.EX (o2r r, o2c c)
    | ALL (r,c) -> CoAlg.AX (o2r r, o2c c)
    | MIN (i,r,c) -> CoAlg.MIN (i, o2r r, o2c c)
    | MAX (i,r,c) -> CoAlg.MAX (i, o2r r, o2c c)


let coalg_global_assumption_of_owl (ax:axiom): CoAlgFormula.formula =
    let o2c = coalg_formula_of_owl in
    match ax with
    | ANNOTATION _ -> CoAlg.TRUE (* ignore annotations *)
    | DECLARATION _ -> CoAlg.TRUE (* ignore declarations *)
    | SUBCLASS (a,b) -> o2c (OR [NOT a; b])
    | EQUIVALENTCLASSES l ->
        (* if a_1 ... a_n are equivalent, then at each point it is
             a_1 ∧ … ∧ a_n (all hold) or ¬a_1 ∧ … ∧ ¬a_n (none holds) *)
        let conjunction = L.fold_left (fun x y -> CoAlg.AND (x, y)) CoAlg.TRUE in
        let allhold = conjunction (L.map o2c l) in
        let nonehold = conjunction (L.map (CoAlg.const_not << o2c) l) in
        CoAlg.OR (allhold, nonehold)
    | DISJOINTCLASSES l -> raise (ConversionError "not implemented yet")
    | DISJOINTUNION (c,l) -> raise (ConversionError "not implemented yet")

let to_coalg (uri,axioms): R.sortTable * R.nomTable * R.assumptions =
  let assumptions = L.map coalg_global_assumption_of_owl axioms in
  (* make assumptions of type CoAlgFormula.formula *)
  let (assumptions, fnctor) = (* ofcourse functor leads to "Error: pattern expected".. *)
    try
      (L.map CoAlg.convertToK assumptions, CM.MultiModalK)
    with CoAlg.ConversionException _ ->
      try (L.map CoAlg.convertToGML assumptions, CM.GML)
      with CoAlg.ConversionException f ->
        raise (ConversionError (CoAlg.string_of_formula f))
  in
  let sorts = [| (fnctor, [0]) |] in
  let nomTable name =
    assert (CoAlgFormula.isNominal name);
    Some 0
  in
  (* make assumptions "sorted" *)
  let assumptions = L.map (fun f -> (0,f)) assumptions in
  (sorts,nomTable,assumptions)

