open OWL
open CoolUtils
module OWLFP = OWLFunctionalParser
module CF = CoAlgFormula
module CR = CoAlgReasoner
module L = List

exception Exception of string

let string_of_stdin (): string =
    let buf = ref "" in
    (try
      while true do
        let input = read_line () in
        buf := !buf ^ input ^ "\n"
      done
    with End_of_file -> ());
    !buf

let owlcat () =
    let content = string_of_stdin () in
    let t = OWLFP.tree_of_string content in
    (*
    let t = OWLFP.Node (L.map OWLFP.tree_of_intree
                            (OWLFP.intree_list_of_tree t)) in
    *)
    let ts = OWLFP.intree_list_of_tree t in
    let show sometree =
        (OWLFP.string_of_intree (OWLFP.string_of_annotated id) sometree)
    in
    let ast:string = String.concat "\n" (L.map show ts) in
    let _ = ast in
    let (ps,os) = OWLFP.parse content in
    foreach_l ps (print_endline << string_of_prefix);
    foreach_l os (print_endline << string_of_ontology)
    (*
    print_endline (OWLFP.string_of_tree (OWLFP.string_of_annotated id) t)
    *)

let consistency () =
    let content = string_of_stdin () in
    (* ignore prefixes and assume that there is exactly one ontology *)
    let (_, ontology) = match OWLFP.parse content with
        | (u, o :: _) -> (u,o)
        | (_, []) -> raise (Exception "File needs to contain at least one ontology")
    in
    let (sorts,nomTable,axioms) = OWL.to_coalg ontology in
    let cons = CR.isSat sorts nomTable axioms (0,CF.TRUE) in
    let cons = if cons then "yes" else "no" in
    let (ontoname, _) = ontology in
    Printf.printf "%s consistent: %s\n" ontoname cons


let rec
    printUsage () =
        let p = Printf.printf in
        p "Usage: %s SUBCMD [ARGS...]\n" (Sys.argv.(0)) ;
        p "\n";
        p "Where SUBCMD is one of:\n";
        foreach_l (tmp_subcmds ()) (fun (name,_,desc) ->
            Printf.printf "    %-10s %s\n" name desc
        );
        p "ARGS is subcommand specific.\n"
and
    tmp_subcmds () =
        let s name func description = (name,func,description) in
        [ s "cat" owlcat "Reads a functional style OWL file from stdin and prints it"
        ; s "consist" consistency "Checks the consistency of a OWL file from stdin"
        ; s "-h"    printUsage "Prints this help screen"
        ; s "help"  printUsage "Prints this help screen"
        ]

let subcmds = tmp_subcmds ()

let _ =
  if Array.length Sys.argv < 2 then printUsage()
  else
    let cmd = Sys.argv.(1) in
    try (
      let (_,f,_) = List.find (fun (n,_,_) -> (n = cmd)) subcmds in
      try f ()
      with | OWLFP.ParseError (a,str) -> (
                print_endline ((OWLFP.string_of_annotation a)^": "^str);
                exit 1
                )
    ) with Not_found -> (
        Printf.printf "Error: Subcommand %s not found.\n" cmd;
        printUsage ()
    )


